//
//	File:			Krislet.java
//	Author:		Krzysztof Langner
//	Date:			1997/04/28
//

//	Modified by:		Tarek Hassan 
//	File:			      IAS.java
//	Date:			      04 July 2001

//  Modified by:  Paul Marlow - January 2002

import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.UIManager;
import java.awt.*; 
import javax.swing.*;  

//***************************************************************************
//
//	This is main object class
//  IAS = In the Agent's Shoes
//
//***************************************************************************

class IAS implements SendCommand
{
  private DatagramSocket m_socket;  // Socket to communicate with server
  private InetAddress m_host;       // Server address
  private int m_port;	            // server port
  private String m_team;		// team name
  private SensorInput m_brain;	// input for sensor information
  private boolean m_goalie;         // indicates if player is the goalie
  private boolean packFrame = false;
  private PlayerFrame p_frame;
  private PlayerCanvas p_canvas;
  private Vector commands;
  private Command currentCommand;
  private char side;
  private double stamina = 2000;    // server default value assumed
  private float time = 0;

  // constants
  private static final int MSG_SIZE = 2048;	// Size of socket buffer

  //===========================================================================
  // Initialization member functions

  //---------------------------------------------------------------------------
  // The main appllication function.
  // Command line format:
  //
  // IAS [-parameter value]
  //
  // Parameters:
  //
  //	host (default "localhost")
  //		The host name can either be a machine name, such as "java.sun.com"
  //		or a string representing its IP address, such as "206.26.48.100."
  //
  //	port (default 6000)
  //		Port number for communication with server
  //
  //	team (default IAS)
  //		Team name. This name can not contain spaces.

  public static void main(String a[]) throws SocketException, IOException
  {
    String	hostName = new String("");
    String	team = new String("GrayMatterUtd");
    int	port = 6000;
    boolean goalie = false;

    try
    {
      // First look for parameters
	for( int c = 0 ; c < a.length ; c += 2 )
	{
	  if( a[c].compareTo("-host") == 0 )
	  {
	    hostName = a[c+1];
	  }
	  else if( a[c].compareTo("-port") == 0 )
	  {
	    port = Integer.parseInt(a[c+1]);
	  }
	  else if( a[c].compareTo("-team") == 0 )
	  {
	    team = a[c+1];
	  }
	  else
	  {
	    throw new Exception();
	  }
	}
    }
    catch(Exception e)
    {
	System.err.println("");
	System.err.println("USAGE: IAS [-parameter value]");
	System.err.println("");
	System.err.println("    Parameters  value          default");
	System.err.println("   --------------------------------------");
	System.err.println("    host        host_name      localhost");
	System.err.println("    port        port_number    6000");
	System.err.println("    team        team_name      GrayMatterUtd");
	System.err.println("");
	System.err.println("    Example:");
	System.err.println("      IAS -host www.host.com -port 6000 -team GrayMatterUtd");
	return;
    }

    try  {
      UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
    }
    catch(Exception e) {
    }

    IAS player = new IAS(InetAddress.getByName(hostName), port, team, goalie);

    // enter main loop
    player.mainLoop();
  }

  //---------------------------------------------------------------------------
  // This constructor opens socket for  connection with server
  public IAS(InetAddress host, int port, String team, boolean goalie)
	throws SocketException
  {
    m_socket = new DatagramSocket();
    m_host = host;
    m_port = port;
    m_team = team;
    m_goalie = goalie;

    p_frame = new PlayerFrame();
    //Validate frames that have preset sizes
    //Pack frames that have useful preferred size info, e.g. from their layout
    if (packFrame)
      p_frame.pack();
    else
      p_frame.validate();

    //Center the window
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    Dimension frameSize = p_frame.getSize();
    if (frameSize.height > screenSize.height)
      frameSize.height = screenSize.height;
    if (frameSize.width > screenSize.width)
      frameSize.width = screenSize.width;
    p_frame.setLocation((screenSize.width - frameSize.width) / 2, (screenSize.height - frameSize.height) / 2);
    p_frame.setVisible(true);
    p_canvas = p_frame.p_canvas;
    p_canvas.setTeam(team);
  }

  //---------------------------------------------------------------------------
  // This destructor closes socket to server
  public void finalize()
  {
    m_socket.close();
  }


  //===========================================================================
  // Protected member functions

  //---------------------------------------------------------------------------
  // This is main loop for player
  protected void mainLoop() throws IOException
  {
    byte[] buffer = new byte[MSG_SIZE];
    DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE);
    // first we need to initialize connection with server
    init();

    m_socket.receive(packet);
    parseInitCommand(new String(buffer,0));
    m_port = packet.getPort();

    // Now we should be connected to the server
    // and we know side, player number and play mode
    while( true )
      parseSensorInformation(receive());
   }


  //===========================================================================
  // Implementation of SendCommand Interface

  //---------------------------------------------------------------------------
  // This function sends move command to the server
  public void move(double x, double y)
  {
    send("(move " + Double.toString(x) + " " + Double.toString(y) + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends turn command to the server
  public void turn(double moment)
  {
    send("(turn " + Double.toString(moment) + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends turn command to the server
  public void turnNeck(double angle)
  {
    send("(turn_neck " + Double.toString(angle) + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends dash command to the server
  public void dash(double power)
  {
    send("(dash " + Double.toString(power) + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends catch command to the server
  public void catchBall(double direction)
  {
    send("(catch " + Double.toString(direction) + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends kick command to the server
  public void kick(double power, double direction)
  {
    send("(kick " + Double.toString(power) + " " + Double.toString(direction) + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends say command to the server
  public void say(String message)
  {
    send("(say " + message + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends chage_view command to the server
  public void changeView(String angle, String quality)
  {
    send("(change_view " + angle + " " + quality + ")");
  }

  //---------------------------------------------------------------------------
  // This function sends sense_body command to the server
  public void sensebody()
  {
    send("(sense_body)");
  }

  //---------------------------------------------------------------------------
  // This function parses initial message from the server
  protected void parseInitCommand(String message) throws IOException
  {
    StringTokenizer tokenizer = new StringTokenizer(message,"() ");

    // We need init token
    if( tokenizer.nextToken().compareTo("init") != 0)
    {
      throw new IOException(message);
    }

    // Read tokenized string
    side = tokenizer.nextToken().charAt(0);
    p_canvas.setSide(side);
    String playerNum = tokenizer.nextToken();
    String playMode = tokenizer.nextToken();

    // initialize player's brain with following parameters:
    // 1st = Player which the new Brain corresponds to
    // 2nd = Name of the team the player belongs to
    // 3rd = Side of playing field - l = left, r = right
    // 4th = Player number for the specified side - 1st player = 1, 2nd = 2 etc.
    // 5th = state of game - initial state = "before_kick_off"

    // Create regular player
    m_brain = new Brain(this, m_team, side, Integer.parseInt(playerNum), playMode);
  }


  //===========================================================================
  // Here comes collection of communication function

  //---------------------------------------------------------------------------
  // This function sends initialization command to the server
  private void init()
  {
    if (!m_goalie)
    {
      send("(init " + m_team + " (version 4.00))");
    }
    else
    {
      send("(init " + m_team + " (version 4.00) (goalie))");
    }
  }

  //---------------------------------------------------------------------------
  // This function parses sensor information
  private void parseSensorInformation(String message)
  {
    StringTokenizer tokenizer = new StringTokenizer(message,"() ");
    String sensor_info = tokenizer.nextToken();
    // First check kind of information
    if (sensor_info.compareTo("see") == 0) // seE information
    {
      //System.out.println(message.trim());
      VisualInfo info = new VisualInfo(message);
      info.parse();
      m_brain.see(info);
      p_canvas.setObjects(info.visuals);
      p_canvas.repaint();
    }
    else if (sensor_info.compareTo("sense_body") == 0) // seNse information
    {
      SenseBodyInfo senseInfo = new SenseBodyInfo(message);
      m_brain.sense(senseInfo);
      stamina = senseInfo.stamina;
      time = senseInfo.time;
    }
    else if (sensor_info.compareTo("hear") == 0) // heAr information
    {
      parseHear(message);
    }
  }

  //---------------------------------------------------------------------------
  // This function parses hear information
  private void parseHear(String message)
  {
    // get hear information
    StringTokenizer tokenizer = new StringTokenizer(message,"() ");
    int time;
    String sender;

    // skip hear token
    tokenizer.nextToken();
    time = Integer.parseInt( tokenizer.nextToken() );
    sender = tokenizer.nextToken();
    if (sender.compareTo("referee") == 0)
    {
      String hear_message = tokenizer.nextToken();
      p_frame.gameMode.setText(hear_message);
      m_brain.hear(time, tokenizer.nextToken());
    }
    else if (sender.compareTo("self") != 0)
      m_brain.hear(time, Integer.parseInt(sender), tokenizer.nextToken());
  }


  //---------------------------------------------------------------------------
  // This function sends via socket message to the server
  private void send(String message)
  {
    byte[] buffer = new byte[MSG_SIZE];
    message.getBytes(0, message.length(), buffer, 0);

    DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE, m_host, m_port);

    try {
      m_socket.send(packet);
    }catch(IOException e){
      System.err.println("socket sending error " + e);
    }
  }

  //---------------------------------------------------------------------------
  // This function waits for new message from server
  private String receive()
  {
    byte[] buffer = new byte[MSG_SIZE];
    DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE);
    try{
      m_socket.receive(packet);
    }catch(IOException e){
      System.err.println("socket receiving error " + e);
    }
    return new String(buffer,0);
  }

  //---------------------------------------------------------------------------
  // gets and removes the next command from the commands queue and sends it to the server
  public void nextCommand()
  {
    p_frame.setStamina(stamina);
    p_frame.setTime(time);
    commands = p_canvas.control.m_commands;
    if (!commands.isEmpty()){
      currentCommand = (Command) commands.remove(0);
      String type = currentCommand.getType();
      if (type.equals("dash")){
        dash(currentCommand.getVar1());
        p_frame.setStatusBar(type + " with power = " + (int)currentCommand.getVar1());
        if (p_canvas.control.running){    // 'running' is true while dash button is held down
          commands.insertElementAt(currentCommand, 0);
        }
      }else if (type.equals("turn")){
        turn(currentCommand.getVar1());
        p_frame.setStatusBar(type + " with moment = " + (int)currentCommand.getVar1());
      }else if (type.equals("kick")){
        kick(currentCommand.getVar1() , currentCommand.getVar2());
        p_frame.setStatusBar(type + " with power = " + (int)currentCommand.getVar1() + " & angle = " + (int)currentCommand.getVar2());
      }
    }
  }
}
